function [f,z,tol] = dist_uncont_hybrid(A, B, options)
% Copyright Ming Gu, Emre Mengi, Michael Overton, Jianlin Xia, Jiang Zhu
% Updated November 4th 2005
% call : [f,z] = dist_uncont_hybrid(A,B,options)
% task : Computes the distance to uncontrollability
%        of the system associated with the matrices
%        (A,B) using the method in [2] or in [5]. At the 
%        termination f differs from the distance to 
%        uncontrollability by at most tol.
%
%        The first order continuous control system with 
%        a single input
%             (*)  x' = Ax + Bu
%        is called controllable if for any given pair
%        of a initial state vector and a desired state
%        vector, there exists an input that can alter
%        the initial state to the desired state in finite
%        time. Assume (*) is controllable, then the
%        distance to uncontrollability of (*) is defined
%        as
%             (**)  tau(A,B) = min(||delA, delB||: the system
%                                 (A+delA,B+delB) is uncontrollable)
%        or, as shown by Eising [1], equivalently
%             (***) tau(A,B) = min_lambda sigma_min([A-lambda I, B]).
%
%        Gu in [3], introduced an algorithm to compute tau(A,B)
%        within a factor of two. The algorithm is based on the
%        test
%                   tau(A,B) <= delta1
%        or
%                   delta2 < tau(A,B).
%        Let delta be an estimate for tau(A,B), at each iteration 
%        Gu's algorithm performs the test above with 
%        delta1=delta and delta2=delta/2. Either the first 
%        condition holds, therefore we can refine the bounds 
%        by replacing delta by delta/2 and repeat the test 
%        above or the second condition holds and we can terminate 
%        with a delta satisfying  tau(A,B) in (delta/2, 2*delta] 
%        (since we know tau(A,B) < 2*delta from the previous
%        iteration). To perform the tests we need to compute 
%        the real eigenvalues of matrices formed by Kronecker 
%        products which requires O(n^6) floating point operations.
%
%        Burke, Lewis and Overton later came up with a trisection
%        algorithm [4] that keeps a lower bound L and an 
%        upper bound U on the distance to uncontrollability. In 
%        the test above delta1 is set L+2*(U-L)/3 and delta2 is   
%        set L+(U-L)/3. If delta2 < tau(A,B) is verfied then lower 
%        bound L is updated to L+(U-L)/3, otherwise upper bound
%        bound is updated to L+2*(U-L)/3. In this manner, it is
%        possible to obtain the distance to uncontrollability
%        to any desired accuracy.
%
%        In [2] more efficient methods to perform the above
%        test are discussed. The eigenvalue problems are
%        simplified, but more importantly, since we only
%        need real eigenvalues of matrices formed by Kronecker
%        products, they can be solved more efficiently by
%        means of inverse iterations by choosing real shifts  
%        appropriately. Let K be the matrix with Kronecker
%        product matrices and u be a vector, the crucial point
%        is that (K-sigma I)^-1 u can be computed by solving a
%        Sylvester equation which requires only O(n^3) operations.
%        We call the method in [2] new verification scheme.
%
%        In [5] an alternative way of verification is presented.      
%        This method does not require the solution of eigenvalue
%        problems involving Kronecker products. Instead a bunch
%        of eigenvalue problems that are twice the size of the
%        input are solved. The number of eigenvalue problems
%        that needs to be solved increases as the lower and
%        upper bounds get close to each other. This method is
%        suggested only for a few digits of precision. We call
%        this method fastest verification scheme (but bear in
%        mind that this is the case only for low precision).
%
%        With this implementation the distance to uncontrollability
%        can be computed using trisection as described above
%        but with the new verification scheme by setting method=0.
%        When method=1, BFGS is used to locate a local minimum
%        of sigma_min([A-lambda I, B]). To test whether the
%        computed minumum eta is indeed a global minimum, it
%        applies the new verification scheme with real
%        eigenvalue extraction techniques (when eig_method=1,
%        otherwise it uses standard Matlab routine eig) by 
%        setting delta1 = eta - 0.5*tol and delta2 = eta - tol.
%        It terminates if tau(A,B) > delta2 is verified,
%        otherwise BFGS is used to locate another local
%        minimum with a starting point alpha+beta i that
%        is returned by the new verification scheme such that
%        sigma_min([A-(alpha+beta i) I, B]) <= delta1. Finally
%        when method=2 fastest verification scheme is applied
%        (the value of eig_method is irrelavent with the fastest
%        method).
%
%        References:
%        [1] R. Eising, Between controllable and uncontrollable,
%            Syst. Contr. Lett., 4(5)(1984), pp. 263-264.
%        [2] M.Gu, E. Mengi, M.L. Overton, J. Xia and J. Zhu, Fast
%            Methods for Estimating the Distance to
%            Uncontrollability, Manuscript.
%        [3] M. Gu, New Methods for Estimating the Distance to
%            Uncontrollability, SIAM J. Matrix Anal. Appl.,
%            21(3) (2000), pp. 989-1003.
%        [4] J.V. Burke, A.S. Lewis and M.L. Overton, Pseudospectral
%            Components and the Distance to Uncontrollability,
%            SIAM J. Matrix Anal. Appl. 26 (2004), pp. 350-361.
%        [5] E. Mengi, An algorithm for estimating the distance to
%            uncontrollability with a few digits of precision,
%            in preperation.
%
% input:
%           A          : nxn coefficient matrix for the state 
%                        vectors.
%           B          : nxm coefficient matrix for the input.
%                        m must be smaller than n.
%           options    : structure with the fields (If any of the 
%                        fields is omitted or this argument is not
%                        provided then the default value mentioned
%                        below is used).
%                        method:
%                           0 : use trisection with the new scheme.
%                           1 : use BFGS to locate local minimums
%                               and check whether a local minimum
%                               is the global minimum by the help
%                               of the new scheme (default).
%                           2 : use trisection but with the fastest
%                               verification scheme (recommended).
%                        eig_method:
%                           the method to be used for the eigenvalue
%                           computations,
%                           0 : use eig of Matlab, rather inefficient
%                               with complexity O(n^6), however 
%                               reliable.
%                           1 : use shifted inverse iterations 
%                               with the divide and conquer approach
%                               (default).
%                        print_dtl:
%                           print the details of the progress of the 
%                           algorithm or not.
%                           0 : don't print (default).
%                           1 : print.
%                        tol:
%                           the absolute accuracy with which the distance
%                           to uncontrollability is computed; expect the 
%                           inequality f - tau <= tol hold (If this field is
%                           not provided a tolerance for which the distance
%                           to uncontrollability can be computed accurately
%                           is estimated as suggested in [2]).
%                        est_norm:
%                           an upper bound on the norm of the Kronecker
%                           product matrix (default: 10^4*||A B||).
%                        eigs_tol:
%                           tolerance that is used by eigs in the
%                           termination criterion (default: 10^-7, set
%                           this default value bigger if you experience
%                           convergence problems with eigs).
%                           
%
%
%
%
% output:
%           f       : computed distance to uncontrollability.
%                     More precisely the exact value is within
%                     a tolerance of tol.
%           z       : the point in the complex plane where f is
%                     achieved. That is sigma_min [A-zI B] = f.
%           tol     : accuracy with which the distance to uncontrollability
%                     is computed.


% error checks
if (nargin < 2)
    error('The matrices A and B must be provided as input')
end

if (nargout > 3)
    error('Wrong number of output arguments')
end 




[sar,sac] = size(A);
[sbr,sbc] = size(B);

if (sar ~= sac)
    error('A must be a square matrix');
end

if (sar ~= sbr)
    error('A and B must have same number of rows');
end

if (sbc > sbr)
    error('B must have at least as many rows as its columns')
end
   




sinvals = svd([A B]);
% set the defaults for the options if they are
% not provided by the user
if (nargin < 3)
    % default uses BFGS
    method = 1;
    % default uses the efficient version for eigenvalue computations
    eig_method = 1;
    % default don't print the details
    print_dtl = 0;
    % default suggest a tolerance to the user for which the computation
    % can be performed accurately
    set_tol = 1;
    tol = 10^-8;
    % default upper bound on the norm of the Kronecker product matrices
    est_normAK = 10^4*max(sinvals);
    % default tolerance for the termination criteria for eigs
    eigs_tol = 10^-7;
else % copy the options provided by the user to the local variables
    
    if isfield(options,'method')        
        % error check
        if options.method ~= 0 & options.method ~= 1 & options.method ~= 2
            error('The method field of options can be either 0, 1 or 2.')
        end
        
        method = options.method;
    else        
        method = 1;
    end
    
    if isfield(options,'eig_method')        
        % error check
        if options.eig_method ~= 0 & options.eig_method ~= 1
            error('The eig_method field of options can be either 0 or 1.')
        end
                
        eig_method = options.eig_method;
    else
        eig_method = 1;
    end
    
    if isfield(options, 'print_dtl')
        % error check
        if options.print_dtl ~= 0 & options.print_dtl ~= 1
            error('The print_dtl field of options can be either of 0 or 1.')
        end

        print_dtl = options.print_dtl;
    else
        print_dtl = 0;
    end
    
    if isfield(options, 'tol')
        % error check
        if ~isnumeric(options.tol) | options.tol < 0
            error('The tol field of options must be a positive real number')
        end
        
        set_tol = 0;
        tol = options.tol;
    else
        set_tol = 1;
        
        if method == 2
            tol = 10^-2;
        else
            if method == 1
                tol = 10^-8;
            else
                tol = 10^-10;
            end
        end
    end
    
    if isfield(options,'est_norm')
        % error check
        if ~isnumeric(options.est_norm) | options.est_norm < 0
            error('The est_norm field of options must be a positive real number')
        end
        
        est_normAK = options.est_norm;
    else
        est_normAK = 10^4*max(sinvals);
    end
    
    if isfield(options,'eigs_tol')
        % error check
        if ~isnumeric(options.eigs_tol) | options.eigs_tol < 0
            error('The eigs_tol field of options must be a positive real number')
        end
        
        eigs_tol = options.eigs_tol;
    else
        eigs_tol = 10^-7;
    end
end


% initialize upper and lower bounds.
% lower bound is meaningless when BFGS is used.
up_bound = min(sinvals);
low_bound = 0;
% z=0 is a point satisfying sigma_min([A-zI B]) = up_bound
% initially.
z = 0;
iter = 0;
total_iter = 0;
check_norm_nomore = 0;




% main loop,
% If trisection is used at each iteration either 
% the upper bound or the lower bound is refined.
%
% If BFGS is used at each iteration a local minimum
% is found and by means of the new verification scheme
% we check whether local minimum is in tol neighborhood
% of the distance to uncontrollability. If tau(A,B) is
% computed with a tolerance tol, we terminate.

% Iterate until the desired accuracy is reached. If the
% user didn't provide the accuracy, make sure that the
% length of the computed interval is shorter than 10^-2.
while up_bound-low_bound > tol
    iter = iter + 1;
    
    if method == 1
        % BFGS to locate a minimum followed by the test
        
        % run BFGS
        pars.nvar = 2;
        pars.fgname = 'msing';
        pars.A = A;
        pars.B = B;
        
        options.x0 = [real(z) imag(z)]';
        options.maxit = 200;
        options.prtlevel = print_dtl;
        
        [zv, up_bound, gnorm, bfgs_iter] = bfgs(pars, options);
        z = zv(1,1)+zv(2,1)*i;
        
        delta1 = max(up_bound - 0.45*tol,tol);
        delta2 = max(up_bound - 0.9*tol,0);
        
        % If the user didn't provide the tolerance, the default
        % tolerance for BFGS is 10^-8. But if this is discovered
        % too small for reliable computation, then it is increased
        % to 10^-2.
        if set_tol & ~check_norm_nomore
            app_norm = approximate_norm(A,B,delta1,2*(delta1-delta2));

            if app_norm > 10^10
                tol = 10^-2;
                check_norm_nomore = 1;
                delta1 = max(up_bound - 0.45*10^-2,10^-2);
                delta2 = max(up_bound - 0.9*10^-2,0);
            end
        end
    else
        % trisection
        delta1 = low_bound + 2*(up_bound - low_bound)/3;
        delta2 = low_bound + (up_bound - low_bound)/3;

        % If the user didn't provide the tolerance, adjust the tolerance
        % to the smallest value such that the distance to uncontrollability
        % can be computed reliably. But also make sure that the tolerance
        % is smaller than 10^-2.
        if method == 0 & set_tol & up_bound - low_bound < 10^-2
            app_norm = approximate_norm(A,B,delta1,2*(delta1-delta2));
            
            if app_norm > 10^10
                tol = 0.99*(up_bound - low_bound);
            end
        end
            
    end

    % main routine to perform the test.
    % first output argument is 1 if delta1 >= tau(A,B)
    %                          0 if delta2 < tau(A,B)
    % second output argument is a z such that
    % sigma_min([A-zI B]) = delta1 provided delta1 >= tau(A,B)
    % holds. Otherwise the second argument is meaningless.
    
    if method == 2
        [test1_test2,zt] = perform_new_test(A,B,delta1,delta2, ...
                                            print_dtl, 100, iter);
        iter2 = 0;
    else        
        [test1_test2,zt,iter2] = perform_test_hybrid(A,B,delta1,delta2, ...
                                            eig_method,print_dtl, ...
                                            est_normAK, eigs_tol, iter);
    end



    total_iter = total_iter + iter2;
  

    if test1_test2
        up_bound = delta1;
        z = zt;                
    else
        low_bound = delta2;
    end    
    
    
    if print_dtl
        fprintf('iter:%d (%3.10f %3.10f] z:%3.10f+%3.10fi\n', iter, low_bound, ...
                                        up_bound, real(z),imag(z));
    end
end


if print_dtl & iter > 0 & eig_method 
    fprintf('average number of calls to eigs: %f\n',total_iter/iter);
end

f = up_bound;
return;
