function [test1_test2,z,iter2] = perform_test_hybrid(...
                                        A,B,delta1,delta2, ...
                                        method,print_dtl, ...
                                        est_normAK, eigs_tol, iter)
% Copyright Ming Gu, Emre Mengi, Michael Overton, Jianlin Xia, Jiang Zhu
% Updated November 4th 2005
% auxiliary function called by dist_uncont
% task : checks which one of
%           (*) tau(A,B) <= delta1 or (**) delta2 < tau(A,B)
%        holds. Note that at least one of the conditions above
%        must be satisfied.
%
% input :
%           A           : nxn complex matrix.
%           B           : nxm complex matrix (m<n).
%           delta1      : the scalar in (*).
%           delta2      : the scalar in (**) (delta2<delta1).
%           method      : the method that will be used to compute
%                         eigenvalues.
%                         0 : Use Matlab's eig.
%                         1 : Use shifted inverse iterations(default).
%           print_dtl   : print the details of the progress of the
%                         algorithm or not.
%                         0 : don't print (default).
%                         1 : print.
%           est_normAK  : An upper bound on the norm of the
%                         Kronecker product matrices.
%           eigs_tol    : tolerance used by eigs for termination.
%           iter        : Number of times this function is called
%                         by the outer function
%
% output :
%           test1_test2 : which one of (*) or (**) holds for the
%                         given delta1 and delta2.
%                         1 : (*) holds.
%                         0 : (**) holds.
%           z           : if (*) holds, this is a point satisfying
%                         sigma_min([A-zI B]) = delta1.
%           iter2       : number of times calls to eigs. 


[sar,sac] = size(A);
[sbr,sbc] = size(B);



% error checks
if (nargin ~= 9)
    error('Wrong number of input arguments');
end

if (nargout ~= 3)
    error('Wrong number of output arguments');
end


if (sar ~= sac)
    error('A must be a square matrix');
end

if (sar ~= sbr)
    error('A and B must have same number of rows');
end

if (sbc > sbr)
    error('B must have at least as many rows as its columns');
end

if (delta2 > delta1)
    error('The upper bound must be greater than the lower bound');
end



% First set delta, eta.
% If the tests below succeed sigma([A-zI B]) = delta
% for some z meaning delta >= tau(A,B). Hence we set
% delta = delta1.
% If the tests fail, then eta can not be in the interval
% [0,2(delta-tau(A,B))], therefore tau(A,B) > delta-eta/2.
% In other words we want to set delta2 = delta-eta/2 or
% eta=2(delta-delta2).
delta = delta1;
eta = 2*(delta-delta2);


% Bhat appears in the matrix formed by Kronecker products 
eyen = eye(sar);
Bhat = (B*B')/delta - delta*eyen;
iter2 = 0;


if ~method
    % This is the naive costly way of computing real eigenvalues
    
    % Perform Kronecker products to obtain the basic matrices
    % appearing in the blocks of the matrix of which real eigenvalues
    % we seek. 
    eyen2 = eye(sar*sar);
    zerosn = zeros(sar);
    zerosn2 = zeros(sar*sar);
    A1 = kron(eyen,A);
    A2 = kron(A-eta*eyen,eyen);
    B1 = kron(eyen,Bhat);
    B2 = kron(Bhat,eyen);
    
    % Then create the 2n^2x2n^2 matrix used in the test.
    AK = ([A1'+A2.' zerosn2; zerosn2 A1+(A2').'] - ...
        [-delta*eyen2 -delta*eyen2; B1 B2.']* ...
        inv([-A1'+(A2').' zerosn2; zerosn2 A1-A2.'])* ...
        [B2.' delta*eyen2; B1 delta*eyen2])/2;




    % use Matlab's eig which means we need a tolerance
    % to decide whether an eigenvalue is pure real.
    % imagtol = 10^-6*sar*sar*(normA+normBhat+delta);
    imagtol = 10^-4;

    eigsAK = eig(AK);
    indx = find(abs(imag(eigsAK)) <= imagtol);
    realeigsAK = real(eigsAK(indx));
    
    
else % else of if ~method
    
    % imag_tol determines how much the imaginary part
    % of a pure real eigenvalue in exact arithmetic can
    % deviate from zero because of rounding errors.
    imag_tol = 10^-4;
    % the largest distance between two distinct eigenvalues
    % that are not distinguishable in finite arithmetic.
    indist_tol = 10^-5;
    
    
    % the real eigenvalues of AK must lie in the interval
    % (-norm(AK),norm(AK)). est_normAK below is an upper
    % bound on the norm of AK.
    
    % initially whole real axis needs to be investigated.
    % In the loop below, the investigated parts of the real
    % axis will be removed. In real_intervals we only keep
    % subintervals of the real axis that still need to be
    % investigated.
    real_intervals(1) = -est_normAK;
    real_intervals(2) = est_normAK;
            
    realeigsAK = [];
    eiglist = [];
    eigcount  = [];

    iter2 = 0;
    while ~isempty(real_intervals) & ...
                length(eiglist) < 2*sar*sar
        low_point = real_intervals(1);
        up_point = real_intervals(2);
        mid_point = (low_point + up_point)/2;        
        
        % apply inverse iteration or arnoldi with the midpoint
        % of an uninvestigated interval as the shift.
        eig_comp = inverse_iteration(A,B,delta,eta,mid_point,eigs_tol);


        % update the real eigenvalue list
        if abs(imag(eig_comp)) < imag_tol & ...
            (isempty(realeigsAK) | ...
                min(abs(realeigsAK - real(eig_comp))) > indist_tol)
           % we found a real eigenvalue not included in the list
           realeigsAK = [realeigsAK real(eig_comp)];
        end
        
        % update the eigenvalue list
        % update the list of uninvestigated subintervals.
        if isempty(eiglist) | ...
            min(abs(eiglist - eig_comp)) > indist_tol
            % this is a new eigenvalue
            eiglist = [eiglist eig_comp];
            eigcount = [eigcount 1];
            
            
            % There is no eigenvalue in the abs(eig_comp-mid_point)
            % neighborhood of mid_point. Current interval is
            % either replaced by two subintervals or discarded.
            
            % add two new subintervals if needed.
            mid_point_low = mid_point - abs(eig_comp-mid_point);
            if mid_point_low > low_point
                num_of_int_times_two = length(real_intervals);
                real_intervals(num_of_int_times_two+1) = low_point;
                real_intervals(num_of_int_times_two+2) = mid_point_low;
            end
            mid_point_up = mid_point + abs(eig_comp-mid_point);
            if mid_point_up < up_point
                num_of_int_times_two = length(real_intervals);
                real_intervals(num_of_int_times_two+1) = mid_point_up;
                real_intervals(num_of_int_times_two+2) = up_point;
            end
            
            
        else
            % eigenvalue was already found
            [mv,indx] = min(abs(eiglist - eig_comp));
            eigcount(indx) = eigcount(indx)+1;
            
            % current interval is simply discarded.
        end
                            
                
        % remove the current interval from the list of
        % uninvestigated subintervals
        % this should be done regardless of a new
        % eigenvalue is found or not.
        num_of_int_times_two = length(real_intervals);
        real_intervals = real_intervals(3:num_of_int_times_two);
        
        iter2 = iter2 + 1;
    end

end % end of else


% check whether AK has any real eigenvalues
if isempty(realeigsAK)
    % test failed
    test1_test2 = 0;
    % in this case content of z does not matter.
    z = 0;
    return;
end % end of if isempty


% AK has real eigenvalues, so check whether
% there exists a real beta such that
% sigma([A-(alpha+beta*i)I B]) =
% sigma([A-(alpha+eta+beta*i)I B]) = delta
% for each real eigenvalue, alpha, of AK.
% Recall that this requires solution of 
% two 2nx2n eigenvalue problems and verification
% whether there is a common pure imaginary
% solution for these problems.

for j = 1 : length(realeigsAK)
    alpha = realeigsAK(j);
            
    realtol = 10^-6;
    smalltol = 10^-2;
    smallertol = 10^-8;
    Aalpha = [-(A'-alpha*eyen) delta*eyen; Bhat (A-alpha*eyen)];
    Aalpha_plus_eta = [-(A'-(alpha+eta)*eyen) delta*eyen;
                                Bhat (A-(alpha+eta)*eyen)];
    eigsAa = eig(Aalpha);
    eigsAae = eig(Aalpha_plus_eta);
    
    
    
    
    
    % extract the imaginary parts of the imaginary
    % eigenvalues of Aalpha and Aalpha_plus_eta.
    indx = find(abs(real(eigsAa)) <= realtol);
    imageigsAa = imag(eigsAa(indx));   
                              
    indx = find(abs(real(eigsAae)) <= realtol);
    imageigsAae = imag(eigsAae(indx));    
    
    
    
        
    % check whether Aalpha and Aalpha_plus_eta have
    % common pure imaginary eigenvalues
    for k = 1 : length(imageigsAa)
        if min(abs(imageigsAae - imageigsAa(k))) < smalltol
            % common imaginary eigenvalue test succeeded
            zp1 = alpha+imageigsAa(k)*i;
            zp2 = alpha+eta+imageigsAa(k)*i;
            
            if min(svd([A-zp1*eyen B])) < delta + smallertol & ...
                 min(svd([A-zp2*eyen B])) < delta + smallertol   
             z = zp1;
             test1_test2 = 1;
             return;
            end
        end
    end
end


% even though the matrix AK have real eigenvalue(s),
% the matrices Aalpha and Aalpha_plus_eta do not
% have a common imaginary eigenvalue for all real
% alpha in Lambda(AK), so the test failed.
test1_test2 = 0;
% content of z does not matter.
z = 0;
return;                                                                                   
