function [nr, z] = numr(A, prtlevel, keyb, plotnum)
% Michael Overton and Emre Mengi (Last updated on March 9th 2005)
%
% call: function [nr, z] = numr(A, prtlevel, keyb, plotnum)
% task
%  Quadratically convergent method to compute the numerical 
%  radius of A.
%
%  DEFINITIONS:
%  The numerical radius of a matrix A is defined as
%
%       (1) r(A) = max{|z| : z in F(A)}
%
%  where F(A) is the field of values of A. As discussed in
%  [1] this is equivalent to the definition,
%
%       (2) r(A) = max_{theta in [0,2pi)} lambda_{max}(H(theta))
%
%  Above lambda_{max} denotes the largest eigenvalue of its argument
%  and H(theta)=(A*e^{i*theta} + A'*e^{-i*theta})/2.
%
%
%  DETAILS OF THE METHOD:
%  The method analogous to the Boyd-Balakrishnan algorithm [2]
%  for H_inf norm is based on definition (2) as described in [3].
%  Let g(theta)=lambda_{max}(H(theta)), we are trying to maximize 
%  g. Each iteration of the method consists of two major steps. 
%  Suppose we are given an estimate of the numerical radius f and 
%  the vector thetamid containing the midpoints of the intervals
%  in which the inequality g(theta) >= f holds. The function 
%  g(theta) is computed at each midpoint. We update f to the maximum 
%  value of g at these midpoints. Next by the help of the theorem below 
%  theta values satisfying
%       (3) f=g(theta)
%  are found. Denote these values theta'=[theta_{1}, theta_{2}, ..., theta_{m}]
%  in sorted order. The updated vector thetamid includes 
%  thetamid_{i} provided g(thetamid_{i}) >= f holds where
%           thetamid_{i} = (theta_{i}+theta_{i+1})/2            if i<m
%                          (theta_{1}+theta_{m} + 2*pi)/2       if i=m
%
%  To find theta satisfying f = g(theta), slightly modified version
%  of theorem 2 in [1] is used, i.e.
%  Theorem : The hermitian matrix H(theta) has f as one of its
%            eigenvalues if and only if the matrix pencil R(f)-lambda*S
%            has the generalized eigenvalue e^{i*theta} where
%            R(f)=[2f*I -A'; I 0] and S=[A 0; 0 I].
%
%
%
%  References:
%  1- Chunyang He and G.A.Watson. An algorithm for computing the
%     numeical radius. IMA Journal of Numerical Analysis, 17:329-342,
%     1997.
%  2- S.Boyd and V.Balakrishnan. A regularity result for the singular
%     values of a transfer matrix and a quadratically convergent 
%     algorithm for computing its L_{inf}-norm. Systems and Control
%     Letters, 15:1-7, 1990.
%  3- E.Mengi and M.L.Overton. Algorithms for the computation of the
%     pseudospectral radius and the numerical radius of a matrix. IMA
%     Journal of Numerical Analysis. To appear.
%
%
% Note     : At each iteration the function g(theta) is computed at
%            the midpoint of each pair of consecutive intersection
%            points. If we have a procedure to determine the multiplicities
%            of the zeros of g(theta)-f, we may instead compute g
%            only at the midpoints where g is known to be greater than f.
%
%
%
% Input:
%       A         : Input matrix(must be square)
%       prtlevel  : 1: Print the details at each iteration
%                   0: Do not print(Default).
%       keyb      : 1: Interact with the user
%                   0: Do not interact(Default).
%       plotnum   : >0: Figure number
%                 : 0: Do not plot(Default).
%
% Output:
%       nr        : Numerical radius of the input matrix A.
%       z         : The point where r(A) is attained in F(A), i.e.
%                   r(A) = |z|.

A = full(A);



if nargin <= 3
    plotnum = 0;
end

if nargin <= 2
	keyb = 0; % default: don't interact with the user
end

if nargin <= 1
	prtlevel = 0; % default: don't print
end




% error checks
if isnaninf(A)|isnaninf([prtlevel keyb])
   error('numr: nan or inf arguments not allowed')
end

n = length(A);
if size(A) ~= [n n]
   error('numr: A must be square')
end





% initializations
thetamid = [0];
% arbitrary values for the first iteration
fold = -.1;
f = 0;
thetabestt = [];
thetabest = thetamid;


fail_horizontal = 0;

iter = 0;
radtol = 10^-6*norm(A);

% main loop
while f > fold & ~fail_horizontal    
	iter = iter + 1;
    
	if prtlevel > 0
        fprintf('numr: iter = %d \n',iter-1);
  		fprintf('numr: f = %22.15e  \n', f);
  		fprintf('numr: thetamid = %22.15e  \n', thetamid);
	end


    % this is a quadratically convergent method, the number of
    % iterations is supposed to be at most 6,7.
	if (iter > 20)
		error('numr: too many iterations')
	end

	thetabestt = thetabest;
	fold = f;
    % compute the function value at each midangle, return
    % the maximum of these function values and the midangle
    % at which the maximum function value is attained
	[f,thetabest] = numr_f(A,thetamid,plotnum,f,iter);
    
    if iter == 1
        f = f + norm(A)*10^-10;
    end
    
    if keyb ~= 0
    	keyboard
	end
    
    % at the first iteration f values are arbitrary
    % update them so that the program enters to the if statement below 
    if (iter == 1)
        fold = f - .1;
    end



	

    % if we have a better estimate compared to the previous iteration
	if (f > fold)
		thetamidold = thetamid;
        % find the zeros of the function h(theta) = g(theta)-alpha
        % return the midpoints of the consecutive pairs of zeros
		thetamid = numr_theta(A,f,thetabest,radtol,prtlevel,plotnum);


		if keyb ~= 0
    			keyboard
		end


        % if thetamid returned by nr_2way_theta is empty, we
        % reached machine accuracy, current f and thetabest
        % are the optimum values and the location where the
        % optimum value is attained.
		fail_horizontal = isempty(thetamid);
	else
        % Otherwise the estimate in the previous iteration is better,
        % so we should have reached machine accuracy in the previous
        % iteration. Restore the values from the previous iteration.
		thetabest = thetabestt;
		f = fold;
	end

end



if isempty(thetabest)
	error('numr: Failed in the first iteration')
end


% set the return variables
nr = f;
% the point where nr is attained is on the boundary of F(A)
% and has the angle -thetabest
z = nr*(cos(thetabest)-i*sin(thetabest));


return;