function [f,z] = poly_dist_uncont(K, B, alpha, c, tol, print_dtl)
% Copyright by Emre Mengi, July 2006
% call: [f,z] = poly_dist_uncont(K, B, alpha, c, tol, print_dtl)
%
% task: Computes the distance to uncontrollability for
%       the mth order system
%
% (1)  alpha_m K_m x^m(t) + alpha_m-1 K_m-1 x^(m-1)(t) + ... + alpha_0 K_0 x(t) = Bu(t)
%
%       where x^j(t) denotes the jth derivative of the state
%       vector x(t) in C^n, u(t) in C^m is the input control.
%       The system (1) is controllable if for any given pair
%       of (x_0,x_f) there exists an input altering x_0 to
%       x_f in finite time.
%
%       Here we assume that the leading coefficient K_m
%       is nonsingular. In this case in [1] Higham and Tisseur
%       has argued that equivalently the system (1) is
%       controllable if and only if the matrix [P(lambda) B]
%       has full rank for all lambda where P(lambda) is the
%       matrix polynomial associated with (1) i.e.
%               P(lambda)=sum_j=0^m lambda^j K_j
%       We define the distance to uncontrollability for the
%       system (1) as
%        
% (2) tau(P,B) = inf{||dK_0 dK_1 ... dK_m dB|| :
%                       (the system (K_0+alpha_0 dK_0,K_1+alpha_1 dK_1,...,K_m+alpha_m dK_m,
%                        B+dB) is uncontrollable}
%
%       that is the norm of the smallest perturbations to the
%       coefficients so that the resulting system is uncontrollable.
%       We have proven an equivalent definition for tau(P,B) (see [3])
%       is
%
% (3) tau(P,B) = inf_lambda sigma_n[P(lambda)/sqrt(s_alpha(lambda)) B]
%
%       for s_alpha(lambda)=sum_j=0^m alpha_j^2|lambda|^2j. The implementation
%       here finds the global minimum of the function
%
% (4)       sigma_min[P(lambda)/sqrt(s_alpha(|lambda|)) B]
%
%       by means of the method described in [3].
%
%
% References:
% 1. Nicholas J. Higham and Francoise Tisseur. More on
%    pseudospectra for polynomial eigenvalue problems and
%    applications in control theory. Linear Alg. Appl.
%    351-352, pp. 435-453. 2002.
% 2. Ming Gu. New methods for estimating the distance to
%    uncontrollability. SIAM J. Matrix Anal. Appl. 21,
%    pp. 989-1003. 2000.
% 3. Emre Mengi. On the estimation of the higher order
%    distance to uncontrollability, submitted to SIMAX.
%
% Input  :
% K         : Sequence of matrices of size nxn. In particular
%             K(:,:,j) must correspond to K_j-1 in (1).
% B         : A tall and skinny matrix of size nxm that is
%             coefficient for the input.
% alpha     : The vector [alpha_0, alpha_1, ..., alpha_k] consisting
%             of nonegative scaling coefficients in (1). 
%             (default [1 1 ... 1])
% c         : An upper bound on 
%                   max 0<=j<=m  r_*^j / sqrt(s_alpha(r_*))
%             where r_*e^itheta_* is a point the higher order distance
%             to uncontrollability is attained. (default c=1.)
% tol       : A tolerance value determining how far the computed
%             value f may lie from the exact tau.
% print_dtl : Print the details or not. 
%             0 (default) : do not print.
%             1           : print.
%
% Output :
% f         : The distance to uncontrollability computed within
%             tol neighborhood of the exact value, i.e.
%             |f-tau| <= tol.
% z         : A point in the complex plane where f is attained,
%             sigma_m[P(z)/sqrt(q(|z|)) B] = f.

if (nargin < 2)
    error('A matrix polynomial and a rectangular matrix must be provided as input')
end

if (nargout > 2)
    error('Wrong number of output arguments')
end


[nr,nc,mp] = size(K);
[br,bc] = size(B);
m = mp - 1;

if (nr ~= nc)
    error('the matrix polynomial should be square');
end

if (nr ~= br)
    error('the matrix polynomial and B must have same number of rows');
end

if (nargin < 6)
    print_dtl = 0; % default: don't print
end
   
if (nargin < 5)
    tol = 10^-8; % default tolerance
end

if (nargin < 4)
    c = 1;
end

if (nargin < 3)
    alpha = ones(1,mp);
else
    [alr,alc] = size(alpha);
    
    if (alr > 1)
        error('alpha must be a row vector')
    end
    
    if (alc ~= mp)
        error('size of alpha must be equal to the order of the system')
    end
end

n = nr;
q = bc;


if (alpha(m+1) > 0)
    U = min(svd([K(:,:,m+1)/alpha(m+1) B]));
else
    sumK = K(:,:,1);
    sumal = alpha(1)^2;
    
    for j = 2:m+1
        sumK = sumK + K(:,:,j);
        sumal = alpha(:,:,j)^2;
    end
    U = min(svd([sumK/sqrt(sumal) B]));
end
    
    
L = 0;
z = 0;

if print_dtl
    fprintf('iter:0 (%3.10f %3.10f] z:0\n', L, U);
end

iter = 1;

maxnorm = norm(K(:,:,2));
for j = 3:m+1
    maxnorm = max(maxnorm,norm(K(:,:,j)));
end

while (U-L > tol)
    
    delta1 = L + 2/3*(U-L);
    delta2 = L + 1/3*(U-L);
    
    delta = delta1;
    eta = (2/m)*acos(1-0.5*((delta1-delta2)/(m*c*maxnorm))^2);
    

    [update_upper,z0] = poly_perform_test(K,B,alpha,delta,eta);

    
    if (update_upper)
        U = delta1;
        z = z0;
    else
        L = delta2;
    end
    
    
    if print_dtl       
       fprintf('iter:%d (%3.10f %3.10f] z:%3.10f+%3.10fi\n', iter, L, ...
                                        U, real(z),imag(z));
    end
    iter = iter + 1;
    
end

f = U;

return;