function wmid = poly_cdi_w(P,gamma,alpha,wwant,realtol,prtlevel)
% Michael Overton and Emre Mengi (Last updated on July 1 2006)
%
% called by poly_cdi.m
%
% Given an estimate alpha of the distance to instability find w
% values satisfying
%       (*)     alpha=f(w)=sigma_min P(w)/p_gamma(|w|) 
% where sigma_min denotes minimum singular value and
%           p_gamma(w) = sqrt(sum_{j=0}^k gamma_j^2 w^2j.
% Let the vector w'=[w(1),w(2),...w(m)] contains the values satisfying 
% (*), the vector wmid containing the midpoints of the consecutive 
% pairs of w' is returned. Formally wmid = [wmid(1),wmid(2),...,wmid(m-1)]
% where for all i, wmid(i) = (w(i) + w(i+1)) / 2. 
%
% Input:
%       P       : Input matrix polynomial(nxnxk).
%       gamma   : Vector of nonnegative scaling factors of size k.
%       alpha   : Current estimate for the distance to instability.
%       wwant   : The location where the current estimate value is
%                 attained.
%       realtol : A tolerance value to accept whether the real part of an
%                 eigenvalue of *-even matrix polynomial is zero or not. 
%                 (See [1] for the definition of *-even polynomial eigenvalue
%                 problem solved.)
%       prtlevel: 1: print the details.
%                 0: do not print.
%
% Output:
%       wmid    : Vector of midpoints, i.e. wmid(i) = (w(i) + w(i+1)) / 2
%                 for 1 <= i <= m-1 where w(1),w(2),...,w(m) are w values
%                 satisfying alpha=f(w).
%
%   
%  REFERENCES
%  [1] E. Mengi. Measures for Robust Stability and Controllability.
%      Ph.D. Thesis. Courant Institute of Mathematical Sciences.
%      New York, August 2006. In preperation.


% get the size of the input matrix and check whether it's square or not.
[nr,nc,kp] = size(P);
if nr ~= nc
    error('poly_cdi_w: P must be square')
end
n = nr;
k = kp-1;


% for the inputs alpha and P create the *-even matrix polynomial.
In = eye(n);
Zn = zeros(n);
smalltol = realtol;

% create the *-even polynomial
for l = 0:2*k
    
    if mod(l,2)
        if l <= k
            Q(:,:,l+1) = [Zn -P(:,:,l+1)'; P(:,:,l+1) Zn];
        else
            Q(:,:,l+1) = zeros(2*n);
        end
    else
        bl = (-1)^(l/2+1)*gamma(l/2+1)^2;
        if l == 0
            Q(:,:,l+1) = [bl*alpha*In P(:,:,l+1)'; P(:,:,l+1) -alpha*In];
        else
            if l <= k
                Q(:,:,l+1) = [bl*alpha*In P(:,:,l+1)'; P(:,:,l+1) Zn];
            else
                Q(:,:,l+1) = [bl*alpha*In Zn; Zn Zn];
            end
        end
    end
end

eQ = mypolyeig(Q);

% keyboard




% extract the eigenvalues whose real parts are close enough to 0 (by the
% real parts close enough to 0 we mean the real parts smaller than the 
% tolerance value).
ind = find(abs(real(eQ)) <= realtol);
eQ = eQ(ind);




if (isempty(eQ)) % check if eQ contains a pure imaginary eigenvalue 
    wmid = [];
else

    % sort eQ according to the imaginary parts
    w = sort(imag(eQ));


    ind2 = [];
    % for each j check if alpha is the smallest singular value of
    % P(w)/p_gamma(|w), previous step guaranteed that alpha is a 
    % singular value of P(w)/p_gamma(w) 
    for j = 1:length(w)
        s = compute_sing(P,gamma,w(j)*i);
    
        if abs(s-alpha) <= smalltol
            ind2 = [ind2; j];
        end
    end


    removed = length(w) - length(ind2);

    % remove the w values for which the matrix A-wiI has the singular value
    % alpha but not the smallest one.
    if removed > 0
        if prtlevel > 0
            fprintf('poly_cdi_w: singular value test removed %d eigenvalues \n', removed);
        end

        w = w(ind2);       
    end
    
    
    % all of the eigenvalues shouldn't be removed by the singular value
    % test. 
    if (isempty(w))
        warning('poly_cdi_w: singular value test removed all of the intersection points')
        wmid = [];
        
        return;
    end


    % There should be even number of intersection pointa  
    if (mod(w,2) == 1)
        error('poly_cdi_w: after singular value test there are odd number of eigenvalues')
    end
    
    
    
    
    
    
    
    % organize in pairs and take midpoints
    ind = 0;

    for j=1:length(w)-1
        wlow = w(j);
        whigh = w(j+1);


        % before taking the midpoint, if this interval is not very short,
        % check and see if wwant is in this interval, well away from the
        % end points.  If so, break this pair into two pairs, one above
        % and one below wwant, and take midpoints of both.
        inttol = .01 * (whigh - wlow);
        if wwant > wlow + inttol & ...
                wwant < whigh - inttol

            % lower midpoint
            ind = ind + 1;
            wmid(ind,1) = (wlow + wwant)/2;
			
            % upper midpoint
            ind = ind + 1;
            wmid(ind,1) = (wwant + whigh)/2;
        else

            % otherwise, if wwant is not in the interval
            % take the midpoint of wlow and whigh
            ind = ind + 1;
            wmid(ind,1) = (wlow + whigh)/2;
        end
    end % end of for j= ...

    
    
end % end of else

return;
