function [beta, theta] = poly_ddi(P, gamma, prtlevel, keyb)
% Emre Mengi and Michael Overton (Last updated on June 30 2006)
%
% call: function [beta, theta] = poly_ddi(P, gamma, prtlevel, keyb)
% task
%  Quadratically convergent method to compute the distance
%  to instability of a discrete system.
%
%
% DEFINITIONS
% Discrete distance to instability of a matrix polynomial is the spectral
% norm of the smallest perturbation to the matrix such that the resulting
% polynomial has an eigenvalue on or outside the unit circle in the
% complex plane, i.e.
%
%       (1)  beta(P,gamma) = min{||DK_0 DK_1 ... DK_K|| : max(|Lambda(P+DP)|) = 1,
%                               where DP(lambda) = sum_j=0^k gamma_j lambda^j DK_j }
%
%  with Lambda(P) denoting the spectrum (set of eigenvalues) of P and |Lambda(P)| 
%  is the vector containing the moduli of the eigenvalues of P. This is equivalent 
%  to the minimization problem
%
%       (2) beta(P,gamma) = min_{theta in [0,2pi)} sigma_{min}(P(e^{itheta})/ ||gamma|| 
%
%  where sigma_{min} denotes the minimum singular value.
%
%  DETAILS OF THE METHOD
%  The method here minimizes g(theta)=sigma_{min}P(e^{i*theta})/||gamma|| over 
%  the unit circle. At each iteration there are two steps. Given a current 
%  estimate of beta(P,gamma), alpha, and the midpoints of the pairs of consecutive 
%  angles satisfying alpha = g(theta), thetamid, we first calculate the function
%  g(theta) at the midpoints. The minimum of the function values replaces 
%  the current estimate of beta(P,gamma). Next the angles satisfying alpha = g(theta) 
%  are found by solving a *-palindromic eigenvalue problem (see [3] for the details). 
%  Let the sorted vector of angles satisfying alpha = g(theta) be theta'. From the 
%  vector theta' the new set of midpoints are computed. The algorithm is a
%  variant of the Boyd Balakrishnan algorithm [1] for H_inf norm which is inspired
%  by the bisection algorithm in [2] for the distance to instability of a matrix.
%
%
%
%  References:
%  1- S.Boyd and V.Balakrishnan. A regularity result for the singular
%     values of a transfer matrix and a quadratically convergent 
%     algorithm for computing its L_{inf}-norm. Systems and Control
%     Letters, 15:1-7, 1990.
%  2- R.Byers. A bisection method for computing the distance of a 
%     stable matrix to the unstable matrices. SIAM Journal on
%     Scientific and Statistical Computing, 9:875-881, 1988.
%  3- E.Mengi. Measures for Robust Stability and Controllability.
%     Ph.D. thesis. Courant Institute of Mathematical Sciences. New
%     York, August 2006. In preperation.
%
%
%
% Input:
%       P         : Input matrix polynomial (C^nxnxk).
%       gamma     : Vector of nonnegative scalars of size k 
%                   (not equal to zero).
%       prtlevel  : 1: Print the details at each iteration
%                   0: Do not print(Default).
%       keyb      : 1: Interact with the user
%                   0: Do not interact(Default).
%
% Output:
%       beta      : Discrete distance to instability of the input 
%                   matrix polynomial P.
%       theta     : The angle where beta(P,gamma) is attained, i.e.
%                   beta(P,gamma) = min_{theta} g(theta).


if nargin <= 3
    keyb = 0; % default: don't interact with the user
end

if nargin <= 2
    prtlevel = 0; % default: don't print
end
% error checks
if nargin < 1
    error('poly_ddi: the input polynomial must be provided')
end

[nr,nc,kp] = size(P);
if nr ~= nc
   error('poly_ddi: P must be square')
end
n = nr;
k = kp-1;



% gamma if provided by user must consist of nonnegative real entries
% not all zero
if nargin >= 2
    [gr,gc] = size(gamma);
    
    if gr ~= 1 | gc ~= kp
        error('poly_ddi: gamma must be row vector of size k')
    end
            
    if sum(gamma-abs(gamma)) ~= 0
        error('poly_ddi: gamma must consist of nonnegative real numbers')
    end
    
    if gamma*gamma' == 0
        error('gamma must be nonzero')
    end
else
    gamma = ones(1,k+1);
end
    


% error checks
if poly_isnaninf(P) | poly_isnaninf([prtlevel keyb]) | ...
        poly_isnaninf(gamma)
   error('poly_ddi: nan or inf arguments not allowed')
end




% check whether P is stable or not. If it is not stable return 0
if max(abs(mypolyeig(P))) >= 1
    beta = 0;
    theta = 0;
    return;
end


% initializations
thetamid = [0];
% arbitrary values for the first iteration
alphaold = 0;
alpha = -.1;
thetabestt = [];
thetabest = thetamid;


fail_horizontal = 0;
iter = 0;

maxnorm = 0;
for j = 0:k
    maxnorm = max(maxnorm,norm(P(:,:,j+1)));
end

radtol = 10^-6*maxnorm;
smalltol = 10^-8*maxnorm;


% main loop
while alpha < alphaold & ~fail_horizontal
    iter = iter + 1;


    % this is a quadratically convergent method, the number of
    % iterations is supposed to be at most 6,7.
    if (iter > 20)
        error('poly_ddi: too many iterations')
    end

    thetabestt = thetabest;
    alphaold = alpha;


    % compute the function value at each midangle, return
    % the minimum of these function values and the midangle
    % at which the minimum function value is attained
    [alpha,thetabest] = poly_ddi_alpha(P,gamma,thetamid);
    
    
    if prtlevel > 0
        fprintf('poly_ddi: thetabest = %22.15e  \n', thetabest);
    end
    
    
    % at the first iteration alpha values are arbitrary
    % update them so that the program enters to the if statement below
    if (iter == 1)
        alphaold = alpha + .1;
    end

	


	

    % if we have a better estimate compared to the previous iteration
    if (alpha < alphaold)
        thetamidold = thetamid;

        % find the zeros of the function h(theta) = g(theta)-alpha
        % return the midangles of the zeros
        thetamid = poly_ddi_theta(P,gamma,alpha,thetabest,radtol,smalltol,prtlevel);

        if keyb ~= 0
            keyboard
        end
        
        % if thetamid returned by ddi_theta is empty, we
        % reached machine precision, current alpha and thetabest
        % are the optimum point and the location where the
        % optimum value is attained.
        fail_horizontal = isempty(thetamid);

    else
    % Otherwise the estimate in the previous iteration is better,
    % so we should have reached machine precision in the previous
    % iteration. Restore the values from the previous iteration.
        thetabest = thetabestt;
        alpha = alphaold;
    end
    
    
    if prtlevel > 0
        fprintf('poly_ddi: iteration %d  \n', iter-1);
        fprintf('poly_ddi: alpha = %22.15e  \n', alpha);
        fprintf('poly_ddi: thetamid = %22.15e  \n', thetamid);
    end
            
end % end of while





if isempty(thetabest)
    error('poly_ddi: failed in the first iteration')
end




% set the return variables, thetabest is already set
beta = alpha;
theta = thetabest;

return;